'''
Created on May 18, 2010

@author: seif
'''

import gobject
import glib
import pango
import gtk
import time
import datetime
from common import event_exists
from ui_utils import *
from model.datasink import SUPPORTED_SOURCES
import gio

pixbuf_cache = {}
no_pixbuf = gtk.icon_theme_get_default().load_icon("gtk-file", 48, 0)

def get_pixbuf_from_cache(gicon, callback, size=48):
	name = gicon.to_string()
	if name in pixbuf_cache:
		callback(pixbuf_cache[name])
	else:
		loadable = gicon.__gtype__.is_a(gio.LoadableIcon)
		if loadable:
			# load asynchronously
			def read_finished(input_stream, res):
				data = input_stream.read_finish(res)
				input_stream.close_async(lambda s, r: s.close_finish(r))

				def idle_cb():
					loader = gtk.gdk.PixbufLoader()
					loader.write(data)
					loader.close()
					pixbuf = loader.get_pixbuf()
					pixbuf_cache[name] = pixbuf
					callback(pixbuf)
					return False

				glib.idle_add(idle_cb)

			def load_finished(arg1, res):
				stream, icon_type = gicon.load_finish(res)
				def idle_cb():
					data = stream.read_async(100*1024, read_finished, glib.PRIORITY_LOW)
					return False

				# we need to finish in idle callback, otherwise this thread
				# will hold global lock and we won't get into main loop
				glib.idle_add(idle_cb)

			gicon.load_async(load_finished)
		else:
			theme = gtk.icon_theme_get_default()
			# blocks: do something with it
			gtk_icon = theme.lookup_by_gicon(gicon, size, size)
			if gtk_icon: icon = gtk_icon.load_icon()
			else: icon = None
			pixbuf_cache[name] = icon
			callback(icon)


def get_icon(item, callback):
	uri = item.uri
	mimetype = item.mimetype
	size = 48
	f = gio.File(uri)
	def callback2( arg1, arg2):
		info = f.query_info_finish(arg2)
		icon_path = info.get_attribute_byte_string(gio.FILE_ATTRIBUTE_THUMBNAIL_PATH)

		def pixbuf_loaded_cb(pixbuf):
			if pixbuf is None: return
			icon = pixbuf.scale_simple(size * pixbuf.get_width() / pixbuf.get_height(),
											size, gtk.gdk.INTERP_BILINEAR)
			callback(icon)

		def add_frame(pixbuf):
			pixbuf_loaded_cb(icon_factory.make_icon_frame(pixbuf, 128))
			
		#print "***", icon
		if icon_path:
			gicon = gio.FileIcon(gio.File(icon_path))
			get_pixbuf_from_cache(gicon, add_frame, 128)
		else:
			gicon = info.get_icon()
			get_pixbuf_from_cache(gicon, pixbuf_loaded_cb)
	
	if f.is_native():
		attr= gio.FILE_ATTRIBUTE_STANDARD_ICON + "," +	gio.FILE_ATTRIBUTE_THUMBNAIL_PATH
		f.query_info_async(attr, callback2)
	else:
		gicon = gio.content_type_get_icon(mimetype)
		icon = get_pixbuf_from_cache(gicon, lambda x: callback(x))

def shade_gdk_color(color, shade):
	"""
	Shades a color by a fraction

	Arguments:
	:param color: a gdk color
	:param shade: fraction by which to shade the color

	:returns: a :class:`Color <gtk.gdk.Color>`
	"""
	f = lambda num: min((num * shade, 65535.0))
	if gtk.pygtk_version >= (2, 16, 0):
		color.red = f(color.red)
		color.green = f(color.green)
		color.blue = f(color.blue)
	else:
		red = int(f(color.red))
		green = int(f(color.green))
		blue = int(f(color.blue))
		color = gtk.gdk.Color(red=red, green=green, blue=blue)
	return color

def match_interpretation(interpretation):
	if interpretation == Interpretation.MUSIC_PIECE:
		return Interpretation.AUDIO
	elif interpretation == Interpretation.SOURCE_CODE:
		return Interpretation.DOCUMENT
	else:
		return interpretation


class Window(gtk.Window):
	'''
	classdocs
	'''

	def __init__(self):
		'''
		Constructor
		'''
		gtk.Window.__init__(self)
		self.set_title("Sezen")
		self.set_default_size(700,400)
		self.connect("destroy", lambda x: gtk.main_quit())
		self.model = None
		self.loaded = False
		self.vbox = gtk.VBox()
		
		self._set_toolbar()
		#self.set_decorated(False)
		#self.maximize()
		self.dview = DashboardView()
		scroll = gtk.ScrolledWindow()
		viewport = gtk.Viewport()
		scroll.add(viewport)
		viewport.add(self.dview)
		viewport.set_shadow_type(gtk.SHADOW_NONE)
		
		def callback(widget, allocation):
			for child in viewport:
				for view in child.views.values():
					view.get_children()[0].set_size_request(1, -1)
		
		self.connect("size-request", callback)
		
		self.toolbar.connect("source-changed", self.dview.set_files)
		scroll.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
		hbox = gtk.HBox()
		self.scroll = scroll
		
		hbox.pack_start(self.toolbar.categorybox ,False, False)
		hbox.pack_start(scroll, True, True)
		self.vbox.pack_start(hbox)
		self.add(self.vbox)
		self.connect("style-set", self.change_style)
		
		self.show_all()
	
	def set_model(self, model):
		self.model = model
		self.toolbar.set_model(model)
		self.model.load_events()
	
	def _set_toolbar(self):
		self.toolbar = Toolbar()
		self.vbox.pack_start(self.toolbar, False, False)
		
	
	def change_style(self, this, old_style):
		style = self.style
		color = style.base[gtk.STATE_NORMAL]
		for child in self.scroll.get_children():
			child.modify_bg(gtk.STATE_NORMAL, color)
		#color = style.bg[gtk.STATE_ACTIVE]
		#self.eventbox.modify_bg(gtk.STATE_NORMAL, color)
			
	
class Toolbar(gtk.VBox):
	__gsignals__ = {
		"source-changed" : (gobject.SIGNAL_RUN_FIRST,
					gobject.TYPE_NONE,
					(gobject.TYPE_PYOBJECT,))
	}
	def __init__(self):
		gtk.VBox.__init__(self)
		self.model = None
		self.sources = None
		self.sorting = 2
		self.toolbar1 = gtk.Toolbar()
		self.toolbar2 = gtk.Toolbar()
		self.vbox = gtk.VBox()
		self.toolbuttons = {}
		self.busy = False
		self.busytoggling = False
		self.active_button = None
		self.results = []
		self.last_search = ""
		self.zoom_level = 90
		self.categorybox = gtk.VBox()
		self.categorybox.set_border_width(3)
		
		self.search = SearchEntry()
		self.search.connect("search", self._set_search)
		self.search.connect("clear", lambda x: self._set_search(None, ""))
		self.pack_start(self.toolbar1, True, True)
	
	def _set_results(self, most):
		self.results = most
		
		results_source_map = {}
		for source in self.sources.keys():
			results_source_map[source] = []
		results_source_map["other"] = []
		t = time.time()
		all = []
				
		keys = SUPPORTED_SOURCES.keys()
				
		for event in most:
			subject = event.subjects[0]
			if event_exists(subject.uri):
				subject.interpretation = match_interpretation(subject.interpretation)
				all.append(event)
				if not subject.interpretation in SUPPORTED_SOURCES.keys():
					results_source_map["other"].append(event)
				else:
					results_source_map[unicode(subject.interpretation)].append(event)
						
			else:
				pass
		
		for key in self.toolbuttons.keys():
			if not key == "all":
				self.toolbuttons[key].set_items(results_source_map[key], self.last_search)
			else:
				self.toolbuttons[key].set_items(all, self.last_search)
		
		if self.active_button:
			self.emit_toggle(self.active_button)
					
	def _set_search(self, widget, search, force = False):
		if self.last_search != search or force:
			self.last_search = search
			self.model.search(search, self._set_results)
			
	def set_model(self, model):
		self.model = model
		self.sources = self.model.get_sources()
		keys = self.sources.keys()
		keys.sort()
		hbox = gtk.HBox()
		
		source = self.sources['all']
		toolbutton = ToolButton(source)
		self.categorybox.pack_start(toolbutton, False, False)
		self.toolbuttons[source["uri"]] = toolbutton
		toolbutton.connect("toggled", self._untoggle_all)
		toolbutton.connect_after("toggled", self.emit_toggle)
		keys.remove("all")
		
		self.categorybox.pack_start(gtk.HSeparator(), False, False, 3)
		
		self.toolbuttons["all"].set_active(True)

		for source in keys:
			source = self.sources[source]
			toolbutton = ToolButton(source)
			self.categorybox.pack_start(toolbutton, False, False)
			self.toolbuttons[source["uri"]] = toolbutton
			toolbutton.connect("toggled", self._untoggle_all)
			toolbutton.connect_after("toggled", self.emit_toggle)
			
		source = {"name": "Other", "icon":"applications-other", "uri": "other"}
		toolbutton = ToolButton(source)
		self.categorybox.pack_start(toolbutton, False, False)
		self.toolbuttons[source["uri"]] = toolbutton
		toolbutton.connect("toggled", self._untoggle_all)
		toolbutton.connect_after("toggled", self.emit_toggle)
		
		self.model.connect("reload", lambda x: self._set_search(None, self.last_search, True))
			
		self.categorybox.show_all()		
		
		vbox = gtk.VBox()
		vbox.pack_start(self.search, True, True, 3)
		hbox = gtk.HBox()
		hbox.pack_start(vbox, True, True, 3)
		
		toolitem = gtk.ToolItem()
		toolitem.add(hbox)
		toolitem.set_expand(True)
		
		self.toolbar1.add(toolitem)
		
		#vbox.pack_start(self.categorybox )
		#vbox = gtk.VBox()
		#vbox.pack_start(hbox, True, True, 3)
		#self.toolbar1.add(vbox)
		
		self.show_all()
		
	def change_zoom(self, widget, level):
		self.zoom_level = level
		self._set_results(self.results)
		self.emit_toggle(self.active_button)
			
	def emit_toggle(self, btn):
		if not self.busy:
			t = time.time()
			items = []
			for i in btn.results:
				items.append(i)
			self.emit("source-changed", items)
	
	def _untoggle_all(self, btn):
		if not self.busy:
			if btn != self.active_button:
				self.busy = True
				for b in self.toolbuttons.values():
					if not b == btn: 
						b.set_active(False)
				self.busy = False
			else:
				btn.set_active(True)
			self.active_button = btn	
			
		
	
class ToolButton(gtk.ToggleButton):
	def __init__(self, source):
		gtk.ToggleButton.__init__(self)
		self.source = source
		self.label = gtk.Label(self.source["name"])
		hbox = gtk.HBox()
		self.add(hbox)
		hbox.pack_end(self.label, True, True, 3)
		self.label.set_alignment(0.0, 0.5)
		if not self.source["name"] == "All":
			icon = icon_factory.load_icon(self.source["icon"], 16, True)
			#icon = icon_factory.greyscale(icon)
			icon = gtk.image_new_from_pixbuf(icon)
			hbox.pack_start(icon, False, False)
		else:
			icon = gtk.gdk.pixbuf_new_from_file_at_size("src/view/zeitgeist-logo.svg", 16, 16)
			#icon = icon_factory.greyscale(icon)
			icon = gtk.image_new_from_pixbuf(icon)
			hbox.pack_start(icon, False, False)
		self.set_relief(gtk.RELIEF_NONE)
		self.results = []
		
	def set_items(self, results, txt):
		self.results = results
		if len(results) ==	0 or txt.strip() == "":
			self.label.set_text(self.source["name"])
			if len(results) == 0:
				self.set_sensitive(False)
			else:
				self.set_sensitive(True)
		else:
			self.set_sensitive(True)


class IconView(gtk.IconView):
	def __init__(self):
		gtk.IconView.__init__(self)
		self.model = gtk.ListStore(gtk.gdk.Pixbuf, str, gobject.TYPE_PYOBJECT)
		#self.set_margin(12)
		self.use_cells = isinstance(self, gtk.CellLayout)
		#self.set_spacing(4)
		self.set_selection_mode(gtk.SELECTION_MULTIPLE)
		self.icon_cell = gtk.CellRendererPixbuf()
		self.icon_cell.set_property("xalign", 0.5)
		self.icon_cell.set_property("yalign", 0.5)
		self.icon_cell.set_fixed_size(128, -1)
		self.pack_start(self.icon_cell, expand=True)
		self.add_attribute(self.icon_cell, "pixbuf", 0)
		self.set_double_buffered(True)
		self.set_reallocate_redraws(True)
		self.set_resize_mode(gtk.RESIZE_IMMEDIATE)
		self.text_cell = gtk.CellRendererText()
		self.text_cell.set_property("wrap-mode", pango.WRAP_WORD_CHAR)
		self.text_cell.set_property("yalign", 0.0)
		self.text_cell.set_property("xalign", 0.5)
		self.text_cell.set_property("ellipsize", pango.ELLIPSIZE_END)
		self.pack_start(self.text_cell, expand=True)
		self.text_cell.set_fixed_size(128, -1)
		#self.set_item_width(128)
		self.add_attribute(self.text_cell, "text", 1)
		self.set_model(self.model)
		self.connect("item-activated", self.launch)
		
	def set_files(self, widget, files):
		self.clear()
		#r = 3
		#for i in range(len(files) / r + 1):
			#thread.start_new_thread(self._set_files, (files[i*r:i*r+r],))
		self._set_files(files)
			
	def _set_files(self, items):
		for item in items:
			self.add_item(item, 48)
		
	def launch(self, view, path):
		item = self.model[path][2]
		launch.launch_uri(item.uri)
		
	def clear(self):
		self.model.clear()
	
	def add_item(self, item, icon_size):
		
		item = item.subjects[0]
		iter = self.model.append([no_pixbuf, item.text, item])
		
		def callback(icon):
			self.model[iter][0] = icon if icon is not None else no_pixbuf
		
		get_icon(item, callback)
	
				
class SearchEntry(gtk.Entry):

	__gsignals__ = {
		"clear" : (gobject.SIGNAL_RUN_FIRST,
				   gobject.TYPE_NONE,
				   ()),
		"search" : (gobject.SIGNAL_RUN_FIRST,
					gobject.TYPE_NONE,
					(gobject.TYPE_STRING,))
	}

	default_text = _("Search...")

	# The font style of the text in the entry.
	#font_style = None

	# TODO: What is this?
	search_timeout = 0

	def __init__(self, accel_group = None):
		gtk.Entry.__init__(self)

		self.set_text(self.default_text)
		self.set_size_request(225, -1)
		self.connect("changed", lambda w: self._queue_search())
		self.connect("focus-in-event", self._entry_focus_in)
		self.connect("focus-out-event", self._entry_focus_out)
		self.connect("icon-press", self._icon_press)
		self.set_icon_from_stock(0, "gtk-find")
		self.set_icon_from_stock(1, "gtk-clear")
		#self.show_all()

	def _icon_press(self, widget, pos, event):
		# Note: GTK_ENTRY_ICON_SECONDARY does not seem to be bound in PyGTK.
		if int(pos) == 1 and not self.get_text() == self.default_text:
			self._entry_clear_no_change_handler()

	def _entry_focus_in(self, widget, x):
		if self.get_text() == self.default_text:
			self.set_text("")
			#self.modify_font(self.font_style)

	def _entry_focus_out(self, widget, x):
		if self.get_text() == "":
			self.set_text(self.default_text)
			#self.modify_font(self.font_style)

	def _entry_clear_no_change_handler(self):
		if not self.get_text() == self.default_text:
			self.set_text("")

	def _queue_search(self):
		if self.search_timeout != 0:
			gobject.source_remove(self.search_timeout)
			self.search_timeout = 0

		if self.get_text() == self.default_text or len(self.get_text()) == 0:
			self.emit("clear")
		else:
			self.search_timeout = gobject.timeout_add(500, self._typing_timeout)

	def _typing_timeout(self):
		if len(self.get_text()) > 0:
			self.emit("search", self.get_text())

		self.search_timeout = 0
		return False
		

class DashboardView(gtk.VBox):

	_zoomlist = [[_("Today"), 1],
				 [_("Yesterday"), 2],
				 [_("Last 7 days"), 7],
				 [_("Last 30 days"), 30],
				 [_("All Time"), 3650]]

	def __init__(self):
		gtk.VBox.__init__(self)
		self.set_border_width(3)
		self.views = {}
		self._zoomlist.reverse()
		self.pack_end(gtk.IconView(), True, True)
		for z in self._zoomlist:
			view = TimeCategory(z)
			self.views[z[1]] = view
			if z[1] <= 7:
				view.set_expanded(True)
			hbox = gtk.HBox()
			hbox.pack_start(view, True, True)
			self.pack_end(hbox, False, False)

			
	def set_files(self, widget, files):
		for view in self.views.values():
			view.treeview.clear()
		today = time.time()
		
		items = {1:[], 2:[], 7:[], 30:[], 3650:[]}
		
		for f in files:
			t= int(f.timestamp)/1000
			offset = int(((today - t)/86400))
			if offset < 1:
				offset = 1
			elif offset < 2:
				offset = 2
			elif offset < 7:
				offset = 7
			elif offset < 30:
				offset = 30
			elif offset < 3650:
				offset = 3650
			else: 
				offset = -1
			if offset > 0:
				items[offset].append(f)
		
		def compare(a, b):
			return cmp(int(b.timestamp), int(a.timestamp)) # compare as integers	
			
		for z in items.keys():
			if z > 7:
				items[z] =  sorted(items[z], compare)
			if len(items[z]) == 0:
				self.views[z].hide_all()
			else:
				self.views[z].show_all()
			self.views[z].set_files(None, items[z])

class TimeCategory(gtk.Expander):
	def __init__(self, zoom):
		gtk.Expander.__init__(self)
		self.zoom = zoom[1]
		self.items = []
		self.label = gtk.Label(zoom[0])
		#self.set_border_width(3)
		self.label.set_markup("<span size='large' underline='low'><b>"+zoom[0]+"</b></span>")
		self.label. set_use_underline(True)
		vbox = gtk.VBox()
		vbox.pack_start(self.label, True, True, 3)
		#vbox.pack_start(gtk.HSeparator(), False, False, 2)
		self.set_label_widget(vbox)
		#self.label.set_padding(20, 3)
		self.treeview = IconView()
		self.add(self.treeview)
		self.connect("style-set", self.change_style)
		self.connect_after("activate", lambda x: self.set_files(None, None))
		self.show_all()
	
	def change_style(self, this, old_style):
		style = self.style
		color = style.bg[gtk.STATE_NORMAL]
		self.label.modify_fg(gtk.STATE_PRELIGHT, color)
		
	def set_files(self, widget, files):
		if files:
			self.items = files
		if self.get_expanded():
			self.treeview.set_files(None, self.items)
			self.treeview.show_all()
		
		

